@@ -2,7 +2,7 @@
 
     MediaTomb - http://www.mediatomb.cc/
 
-    upnp_cds.cc - this file is part of MediaTomb.
+    cont_dir_service.cc - this file is part of MediaTomb.
 
     Copyright (C) 2005 Gena Batyan <bgeradz@mediatomb.cc>,
                        Sergey 'Jin' Bostandzhyan <jin@mediatomb.cc>
@@ -11,7 +11,7 @@
                             Sergey 'Jin' Bostandzhyan <jin@mediatomb.cc>,
                             Leonhard Wimmer <leo@mediatomb.cc>
 
-    Copyright (C) 2016-2024 Gerbera Contributors
+    Copyright (C) 2016-2025 Gerbera Contributors
 
     MediaTomb is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License version 2
@@ -29,8 +29,8 @@
     $Id$
 */
 
-/// \file upnp_cds.cc
-#define LOG_FAC log_facility_t::cds
+/// \file cont_dir_service.cc
+#define GRB_LOG_FAC GrbLogFacility::cds
 
 #include "cont_dir_service.h" // API
 
@@ -39,26 +39,43 @@
 #include "action_request.h"
 #include "cds/cds_container.h"
 #include "cds/cds_item.h"
-#include "config/config_manager.h"
+#include "config/config.h"
+#include "config/config_val.h"
+#include "context.h"
 #include "database/database.h"
+#include "database/db_param.h"
 #include "database/sql_database.h"
+#include "exceptions.h"
 #include "subscription_request.h"
 #include "upnp/clients.h"
+#include "upnp/compat.h"
 #include "upnp/quirks.h"
 #include "upnp/xml_builder.h"
 #include "util/tools.h"
 
 ContentDirectoryService::ContentDirectoryService(const std::shared_ptr<Context>& context,
-    std::shared_ptr<UpnpXMLBuilder> xmlBuilder, UpnpDevice_Handle deviceHandle, int stringLimit)
-    : stringLimit(stringLimit)
-    , config(context->getConfig())
+    const std::shared_ptr<UpnpXMLBuilder>& xmlBuilder, UpnpDevice_Handle deviceHandle,
+    int stringLimit, bool offline)
+    : UpnpService(context->getConfig(), xmlBuilder, deviceHandle, UPNP_DESC_CDS_SERVICE_ID, offline)
+    , stringLimit(stringLimit)
     , database(context->getDatabase())
-    , deviceHandle(deviceHandle)
-    , xmlBuilder(std::move(xmlBuilder))
 {
-    titleSegments = this->config->getArrayOption(CFG_UPNP_SEARCH_ITEM_SEGMENTS);
-    resultSeparator = this->config->getOption(CFG_UPNP_SEARCH_SEPARATOR);
-    searchableContainers = this->config->getBoolOption(CFG_UPNP_SEARCH_CONTAINER_FLAG);
+    actionMap = {
+        { "Browse", [this](ActionRequest& r) { doBrowse(r); } },
+        { "GetSearchCapabilities", [this](ActionRequest& r) { doGetSearchCapabilities(r); } },
+        { "GetSortCapabilities", [this](ActionRequest& r) { doGetSortCapabilities(r); } },
+        { "GetSystemUpdateID", [this](ActionRequest& r) { doGetSystemUpdateID(r); } },
+        { "GetFeatureList", [this](ActionRequest& r) { doGetFeatureList(r); } },
+        { "GetSortExtensionCapabilities", [this](ActionRequest& r) { doGetSortExtensionCapabilities(r); } },
+        { "Search", [this](ActionRequest& r) { doSearch(r); } },
+        { "X_SetBookmark", [this](ActionRequest& r) { doSamsungBookmark(r); } },
+        { "X_GetFeatureList", [this](ActionRequest& r) { doSamsungFeatureList(r); } },
+        { "X_GetObjectIDfromIndex", [this](ActionRequest& r) { doSamsungGetObjectIDfromIndex(r); } },
+        { "X_GetIndexfromRID", [this](ActionRequest& r) { doSamsungGetIndexfromRID(r); } },
+    };
+    titleSegments = this->config->getArrayOption(ConfigVal::UPNP_SEARCH_ITEM_SEGMENTS);
+    resultSeparator = this->config->getOption(ConfigVal::UPNP_SEARCH_SEPARATOR);
+    searchableContainers = this->config->getBoolOption(ConfigVal::UPNP_SEARCH_CONTAINER_FLAG);
 }
 
 void ContentDirectoryService::doBrowse(ActionRequest& request)
@@ -69,63 +86,76 @@ void ContentDirectoryService::doBrowse(ActionRequest& request)
     auto reqRoot = req->document_element();
 
 #ifdef GRBDEBUG
-    if (GrbLogger::Logger.isDebugging(LOG_FAC))
+    if (GrbLogger::Logger.isDebugging(GRB_LOG_FAC))
         for (auto&& child : reqRoot.children()) {
             log_debug("request {} = {}", child.name(), reqRoot.child(child.name()).text().as_string());
         }
 #endif
 
+    // prepare browse parameters
     std::string objID = reqRoot.child("ObjectID").text().as_string();
     std::string browseFlag = reqRoot.child("BrowseFlag").text().as_string();
-    // std::string filter; // not yet supported
-    std::string startingIndex = reqRoot.child("StartingIndex").text().as_string();
-    std::string requestedCount = reqRoot.child("RequestedCount").text().as_string();
+    int startingIndex = stoiString(reqRoot.child("StartingIndex").text().as_string());
+    std::string filter = reqRoot.child("Filter").text().as_string();
+    int requestedCount = stoiString(reqRoot.child("RequestedCount").text().as_string());
     std::string sortCriteria = reqRoot.child("SortCriteria").text().as_string();
 
-    log_debug("Browse received parameters: ObjectID [{}] BrowseFlag [{}] StartingIndex [{}] RequestedCount [{}] SortCriteria [{}]",
-        objID, browseFlag, startingIndex, requestedCount, sortCriteria);
+    log_debug("Browse received parameters: ObjectID [{}] BrowseFlag [{}] StartingIndex [{}] Filter [{}] RequestedCount [{}] SortCriteria [{}]",
+        objID, browseFlag, startingIndex, filter, requestedCount, sortCriteria);
 
     if (objID.empty())
         throw UpnpException(UPNP_E_NO_SUCH_ID, "empty object id");
 
     auto&& quirks = request.getQuirks();
-    auto arr = quirks->getSamsungFeatureRoot(database, objID);
     int objectID = stoiString(objID);
+    auto arr = quirks && objectID == 0 && objID != "0"
+        ? quirks->getSamsungFeatureRoot(database, startingIndex, requestedCount, objID)
+        : std::vector<std::shared_ptr<CdsObject>>();
 
     unsigned int flag = BROWSE_ITEMS | BROWSE_CONTAINERS | BROWSE_EXACT_CHILDCOUNT;
 
     if (browseFlag == "BrowseDirectChildren")
         flag |= BROWSE_DIRECT_CHILDREN;
     else if (browseFlag != "BrowseMetadata")
-        throw UpnpException(UPNP_SOAP_E_INVALID_ARGS,
-            "Invalid browse flag: " + browseFlag);
+        throw UpnpException(UPNP_SOAP_E_INVALID_ARGS, "Invalid browse flag: " + browseFlag);
 
     auto parent = database->loadObject(quirks->getGroup(), objectID);
     auto upnpClass = parent->getClass();
     if (sortCriteria.empty() && (startswith(upnpClass, UPNP_CLASS_MUSIC_ALBUM) || startswith(upnpClass, UPNP_CLASS_PLAYLIST_CONTAINER)))
         flag |= BROWSE_TRACK_SORT;
-
-    if (config->getBoolOption(CFG_SERVER_HIDE_PC_DIRECTORY))
+    if (filter.empty())
+        filter = "*";
+    if (config->getBoolOption(ConfigVal::SERVER_HIDE_PC_DIRECTORY))
         flag |= BROWSE_HIDE_FS_ROOT;
 
     auto param = BrowseParam(parent, flag);
 
-    param.setDynamicContainers(!quirks->checkFlags(QUIRK_FLAG_SAMSUNG_HIDE_DYNAMIC));
-    param.setStartingIndex(stoiString(startingIndex));
-    param.setRequestedCount(stoiString(requestedCount));
+    param.setDynamicContainers(!quirks || !quirks->checkFlags(QUIRK_FLAG_SAMSUNG_HIDE_DYNAMIC));
+    param.setStartingIndex(startingIndex);
+    param.setRequestedCount(requestedCount);
     param.setSortCriteria(trimString(sortCriteria));
     param.setGroup(quirks->getGroup());
+    if (quirks)
+        param.setForbiddenDirectories(quirks->getForbiddenDirectories());
 
+    // Execute database browse
     try {
         if (arr.empty())
             arr = database->browse(param);
         else
             param.setTotalMatches(arr.size());
+    } catch (const SearchParseException& srcEx) {
+        log_warning(srcEx.what());
+        throw UpnpException(UPNP_E_INVALID_ARGUMENT, srcEx.getUserMessage());
+    } catch (const DatabaseException& dbEx) {
+        log_warning(dbEx.what());
+        throw UpnpException(UPNP_E_NO_SUCH_ID, dbEx.getUserMessage());
     } catch (const std::runtime_error& e) {
-        log_error("No such object: {}", e.what());
-        throw UpnpException(UPNP_E_NO_SUCH_ID, "no such object");
+        log_warning(e.what());
+        throw UpnpException(UPNP_E_BAD_REQUEST, UpnpXMLBuilder::encodeEscapes(e.what()));
     }
 
+    // build response
     pugi::xml_document didlLite;
     if (!quirks->blockXmlDeclaration()) {
         auto decl = didlLite.prepend_child(pugi::node_declaration);
@@ -145,7 +175,7 @@ void ContentDirectoryService::doBrowse(ActionRequest& request)
 
     for (auto&& obj : arr) {
         markPlayedItem(obj, obj->getTitle());
-        xmlBuilder->renderObject(obj, stringLimitClient, didlLiteRoot, quirks);
+        xmlBuilder->renderObject(obj, splitString(filter, ','), stringLimitClient, didlLiteRoot, quirks);
     }
 
     std::string didlLiteXml = UpnpXMLBuilder::printXml(didlLite, "", quirks && quirks->needsStrictXml() ? pugi::format_no_escapes : 0);
@@ -170,24 +200,26 @@ void ContentDirectoryService::doSearch(ActionRequest& request)
     auto reqRoot = req->document_element();
 
 #ifdef GRBDEBUG
-    if (GrbLogger::Logger.isDebugging(LOG_FAC))
+    if (GrbLogger::Logger.isDebugging(GRB_LOG_FAC))
         for (auto&& child : reqRoot.children()) {
             log_debug("request {} = {}", child.name(), reqRoot.child(child.name()).text().as_string());
         }
 #endif
 
+    // prepare search parameters
     std::string containerID = reqRoot.child("ContainerID").text().as_string();
     std::string searchCriteria = reqRoot.child("SearchCriteria").text().as_string();
-    std::string startingIndex = reqRoot.child("StartingIndex").text().as_string();
-    std::string requestedCount = reqRoot.child("RequestedCount").text().as_string();
+    int startingIndex = stoiString(reqRoot.child("StartingIndex").text().as_string());
+    std::string filter = reqRoot.child("Filter").text().as_string();
+    int requestedCount = stoiString(reqRoot.child("RequestedCount").text().as_string());
     std::string sortCriteria = reqRoot.child("SortCriteria").text().as_string();
 
-    log_debug("Search received parameters: ContainerID [{}] SearchCriteria [{}] SortCriteria [{}] StartingIndex [{}] RequestedCount [{}]",
-        containerID, searchCriteria, sortCriteria, startingIndex, requestedCount);
+    log_debug("Search received parameters: ContainerID [{}] SearchCriteria [{}] SortCriteria [{}] StartingIndex [{}] Filter [{}] RequestedCount [{}]",
+        containerID, searchCriteria, sortCriteria, startingIndex, filter, requestedCount);
 
     auto&& quirks = request.getQuirks();
     pugi::xml_document didlLite;
-    if (!quirks->blockXmlDeclaration()) {
+    if (!quirks || !quirks->blockXmlDeclaration()) {
         auto decl = didlLite.prepend_child(pugi::node_declaration);
         decl.append_attribute("version") = "1.0";
         decl.append_attribute("encoding") = "UTF-8";
@@ -202,19 +234,30 @@ void ContentDirectoryService::doSearch(ActionRequest& request)
     if (sortCriteria.empty()) {
         sortCriteria = fmt::format("+{}", MetaEnumMapper::getMetaFieldName(MetadataFields::M_TITLE));
     }
-    const auto searchParam = SearchParam(containerID, searchCriteria, sortCriteria,
-        stoiString(startingIndex), stoiString(requestedCount), searchableContainers, quirks->getGroup());
+    if (filter.empty())
+        filter = "*";
+    auto searchParam = SearchParam(containerID, searchCriteria, sortCriteria,
+        startingIndex, requestedCount, searchableContainers, quirks->getGroup());
+    if (quirks)
+        searchParam.setForbiddenDirectories(quirks->getForbiddenDirectories());
 
+    // Execute database search
     std::vector<std::shared_ptr<CdsObject>> results;
-    int numMatches = 0;
     try {
-        results = database->search(searchParam, &numMatches);
-        log_debug("Found {}/{} items", results.size(), numMatches);
+        results = database->search(searchParam);
+        log_debug("Found {}/{} items", results.size(), searchParam.getTotalMatches());
+    } catch (const SearchParseException& srcEx) {
+        log_warning(srcEx.what());
+        throw UpnpException(UPNP_E_INVALID_ARGUMENT, srcEx.getUserMessage());
+    } catch (const DatabaseException& dbEx) {
+        log_warning(dbEx.what());
+        throw UpnpException(UPNP_E_NO_SUCH_ID, dbEx.getUserMessage());
     } catch (const std::runtime_error& e) {
-        log_debug(e.what());
-        throw UpnpException(UPNP_E_NO_SUCH_ID, "no such object");
+        log_warning(e.what());
+        throw UpnpException(UPNP_E_BAD_REQUEST, UpnpXMLBuilder::encodeEscapes(e.what()));
     }
 
+    // build response
     auto stringLimitClient = stringLimit;
     if (quirks->getStringLimit() > -1) {
         stringLimitClient = quirks->getStringLimit();
@@ -222,7 +265,7 @@ void ContentDirectoryService::doSearch(ActionRequest& request)
 
     for (auto&& cdsObject : results) {
         if (!cdsObject->isItem()) {
-            xmlBuilder->renderObject(cdsObject, stringLimitClient, didlLiteRoot);
+            xmlBuilder->renderObject(cdsObject, splitString(filter, ','), stringLimitClient, didlLiteRoot);
             continue;
         }
 
@@ -240,7 +283,7 @@ void ContentDirectoryService::doSearch(ActionRequest& request)
         }
 
         markPlayedItem(cdsObject, title);
-        xmlBuilder->renderObject(cdsObject, stringLimitClient, didlLiteRoot);
+        xmlBuilder->renderObject(cdsObject, splitString(filter, ','), stringLimitClient, didlLiteRoot);
     }
 
     std::string didlLiteXml = UpnpXMLBuilder::printXml(didlLite, "", quirks && quirks->needsStrictXml() ? pugi::format_no_escapes : 0);
@@ -250,7 +293,7 @@ void ContentDirectoryService::doSearch(ActionRequest& request)
     auto respRoot = response->document_element();
     respRoot.append_child("Result").append_child(pugi::node_pcdata).set_value(didlLiteXml.c_str());
     respRoot.append_child("NumberReturned").append_child(pugi::node_pcdata).set_value(fmt::to_string(results.size()).c_str());
-    respRoot.append_child("TotalMatches").append_child(pugi::node_pcdata).set_value(fmt::to_string(numMatches).c_str());
+    respRoot.append_child("TotalMatches").append_child(pugi::node_pcdata).set_value(fmt::to_string(searchParam.getTotalMatches()).c_str());
     respRoot.append_child("UpdateID").append_child(pugi::node_pcdata).set_value(fmt::to_string(systemUpdateID).c_str());
     request.setResponse(std::move(response));
 
@@ -271,14 +314,14 @@ void ContentDirectoryService::markPlayedItem(const std::shared_ptr<CdsObject>& c
 
     auto item = std::static_pointer_cast<CdsItem>(cdsObject);
     auto playStatus = item->getPlayStatus();
-    if (config->getBoolOption(CFG_SERVER_EXTOPTS_MARK_PLAYED_ITEMS_ENABLED) && playStatus && playStatus->getPlayCount() > 0) {
-        std::vector<std::string> markList = config->getArrayOption(CFG_SERVER_EXTOPTS_MARK_PLAYED_ITEMS_CONTENT_LIST);
+    if (config->getBoolOption(ConfigVal::SERVER_EXTOPTS_MARK_PLAYED_ITEMS_ENABLED) && playStatus && playStatus->getPlayCount() > 0) {
+        std::vector<std::string> markList = config->getArrayOption(ConfigVal::SERVER_EXTOPTS_MARK_PLAYED_ITEMS_CONTENT_LIST);
         bool mark = std::any_of(markList.begin(), markList.end(), [&](auto&& i) { return startswith(item->getClass(), markContentMap.at(i)); });
         if (mark) {
-            if (config->getBoolOption(CFG_SERVER_EXTOPTS_MARK_PLAYED_ITEMS_STRING_MODE_PREPEND))
-                title = config->getOption(CFG_SERVER_EXTOPTS_MARK_PLAYED_ITEMS_STRING).append(title);
+            if (config->getBoolOption(ConfigVal::SERVER_EXTOPTS_MARK_PLAYED_ITEMS_STRING_MODE_PREPEND))
+                title = config->getOption(ConfigVal::SERVER_EXTOPTS_MARK_PLAYED_ITEMS_STRING).append(title);
             else
-                title.append(config->getOption(CFG_SERVER_EXTOPTS_MARK_PLAYED_ITEMS_STRING));
+                title.append(config->getOption(ConfigVal::SERVER_EXTOPTS_MARK_PLAYED_ITEMS_STRING));
         }
     }
 
@@ -321,6 +364,9 @@ void ContentDirectoryService::doGetFeatureList(ActionRequest& request) const
     features.append_attribute("xmlns") = "urn:schemas-upnp-org:av:avs";
     features.append_attribute("xmlns:xsi") = "http://www.w3.org/2001/XMLSchema-instance";
     features.append_attribute("xsi:schemaLocation") = "urn:schemas-upnp-org:av:avs http://www.upnp.org/schemas/av/avs.xsd";
+    auto&& quirks = request.getQuirks();
+    if (quirks)
+        quirks->getShortCutList(database, features);
     root.append_child("FeatureList").append_child(pugi::node_pcdata).set_value(UpnpXMLBuilder::printXml(respRoot, "", 0).c_str());
     request.setResponse(std::move(response));
 
@@ -387,42 +433,7 @@ void ContentDirectoryService::doSamsungGetIndexfromRID(ActionRequest& request)
     log_debug("end");
 }
 
-void ContentDirectoryService::processActionRequest(ActionRequest& request)
-{
-    log_debug("start");
-
-    if (request.getActionName() == "Browse") {
-        doBrowse(request);
-    } else if (request.getActionName() == "GetSearchCapabilities") {
-        doGetSearchCapabilities(request);
-    } else if (request.getActionName() == "GetSortCapabilities") {
-        doGetSortCapabilities(request);
-    } else if (request.getActionName() == "GetSystemUpdateID") {
-        doGetSystemUpdateID(request);
-    } else if (request.getActionName() == "GetFeatureList") {
-        doGetFeatureList(request);
-    } else if (request.getActionName() == "GetSortExtensionCapabilities") {
-        doGetSortExtensionCapabilities(request);
-    } else if (request.getActionName() == "Search") {
-        doSearch(request);
-    } else if (request.getActionName() == "X_SetBookmark") {
-        doSamsungBookmark(request);
-    } else if (request.getActionName() == "X_GetFeatureList") {
-        doSamsungFeatureList(request);
-    } else if (request.getActionName() == "X_GetObjectIDfromIndex") {
-        doSamsungGetObjectIDfromIndex(request);
-    } else if (request.getActionName() == "X_GetIndexfromRID") {
-        doSamsungGetIndexfromRID(request);
-    } else {
-        // invalid or unsupported action
-        log_warning("Unrecognized action {}", request.getActionName().c_str());
-        request.setErrorCode(UPNP_E_INVALID_ACTION);
-    }
-
-    log_debug("ContentDirectoryService::processActionRequest: end");
-}
-
-void ContentDirectoryService::processSubscriptionRequest(const SubscriptionRequest& request)
+bool ContentDirectoryService::processSubscriptionRequest(const SubscriptionRequest& request)
 {
     log_debug("start");
 
@@ -435,27 +446,12 @@ void ContentDirectoryService::processSubscriptionRequest(const SubscriptionReque
 
     std::string xml = UpnpXMLBuilder::printXml(*propset, "", 0);
 
-#if defined(USING_NPUPNP)
-    UpnpAcceptSubscriptionXML(
-        deviceHandle, config->getOption(CFG_SERVER_UDN).c_str(),
-        UPNP_DESC_CDS_SERVICE_ID, xml, request.getSubscriptionID().c_str());
-#else
-    IXML_Document* event = nullptr;
-    int err = ixmlParseBufferEx(xml.c_str(), &event);
-    if (err != IXML_SUCCESS) {
-        throw UpnpException(UPNP_E_SUBSCRIPTION_FAILED, "Could not convert property set to ixml");
-    }
-
-    UpnpAcceptSubscriptionExt(deviceHandle,
-        config->getOption(CFG_SERVER_UDN).c_str(),
-        UPNP_DESC_CDS_SERVICE_ID, event, request.getSubscriptionID().c_str());
-
-    ixmlDocument_free(event);
-#endif
-    log_debug("end");
+    return UPNP_E_SUCCESS == GrbUpnpAcceptSubscription( //
+               deviceHandle, config->getOption(ConfigVal::SERVER_UDN), //
+               serviceID, xml, request.getSubscriptionID());
 }
 
-void ContentDirectoryService::sendSubscriptionUpdate(const std::string& containerUpdateIDsCsv)
+bool ContentDirectoryService::sendSubscriptionUpdate(const std::string& containerUpdateIDsCsv)
 {
     log_debug("start {}", containerUpdateIDsCsv);
 
@@ -467,21 +463,7 @@ void ContentDirectoryService::sendSubscriptionUpdate(const std::string& containe
     property.append_child("SystemUpdateID").append_child(pugi::node_pcdata).set_value(fmt::to_string(systemUpdateID).c_str());
 
     std::string xml = UpnpXMLBuilder::printXml(*propset, "", 0);
-
-#if defined(USING_NPUPNP)
-    UpnpNotifyXML(deviceHandle, config->getOption(CFG_SERVER_UDN).c_str(), UPNP_DESC_CDS_SERVICE_ID, xml);
-#else
-    IXML_Document* event = nullptr;
-    int err = ixmlParseBufferEx(xml.c_str(), &event);
-    if (err != IXML_SUCCESS) {
-        /// \todo add another error code
-        throw UpnpException(UPNP_E_SUBSCRIPTION_FAILED, "Could not convert property set to ixml");
-    }
-
-    UpnpNotifyExt(deviceHandle, config->getOption(CFG_SERVER_UDN).c_str(), UPNP_DESC_CDS_SERVICE_ID, event);
-
-    ixmlDocument_free(event);
-#endif
-
-    log_debug("end");
+    return UPNP_E_SUCCESS == GrbUpnpNotify( //
+               deviceHandle, config->getOption(ConfigVal::SERVER_UDN), //
+               serviceID, xml);
 }
